from dishka.dependency_source import (
    Alias,
    ContextVariable,
    Decorator,
    Factory,
)
from dishka.entities.component import Component


class BaseProvider:
    def __init__(self, component: Component | None) -> None:
        if component is not None:
            self.component = component
        self.factories: list[Factory] = []
        self.aliases: list[Alias] = []
        self.decorators: list[Decorator] = []
        self.context_vars: list[ContextVariable] = []


class ProviderWrapper(BaseProvider):
    def __init__(self, component: Component, provider: BaseProvider) -> None:
        super().__init__(component)
        self.factories.extend(provider.factories)
        self.aliases.extend(provider.aliases)
        self.decorators.extend(provider.decorators)



from collections.abc import Iterable
from typing import Any

from dishka.entities.component import DEFAULT_COMPONENT
from dishka.entities.scope import BaseScope
from .base_provider import BaseProvider
from .provider import Provider


def make_root_context_provider(
        providers: Iterable[BaseProvider],
        context: dict[Any, Any] | None,
        scopes: type[BaseScope],
) -> BaseProvider:
    """Automatically add missing `from_context`."""
    # in non-default component, context vars are aliases for it
    # use only those, which declared in Default component provider
    existing_context_vars = {
        var.provides.type_hint
        for provider in providers
        if provider.component == DEFAULT_COMPONENT
        for var in provider.context_vars
    }
    p = Provider()
    if not context:
        return p
    root_scope = next(iter(scopes))
    for type_hint in context:
        if type_hint not in existing_context_vars:
            p.from_context(provides=type_hint, scope=root_scope)
    return p



from dataclasses import dataclass


@dataclass
class ValidationSettings:
    # check if no factory found to override when set override=True
    nothing_overridden: bool = False
    # check if factory is overridden when set override=False
    implicit_override: bool = False
    # check if decorator was not applied to any factory
    nothing_decorated: bool = True


DEFAULT_VALIDATION = ValidationSettings()
STRICT_VALIDATION = ValidationSettings(
    nothing_overridden=True,
    implicit_override=True,
    nothing_decorated=True,
)



from collections.abc import Callable, Sequence
from typing import Any

from dishka.exception_base import DishkaError
from dishka.text_rendering import get_name


class NotAFactoryError(TypeError, DishkaError):
    def __init__(self, attempted_factory: Any) -> None:
        self.attempted = attempted_factory

    def __str__(self) -> str:
        return f"Cannot use {self.attempted!r} as a factory."


class CannotUseProtocolError(NotAFactoryError):
    def __str__(self) -> str:
        return (
            f"Cannot use {self.attempted} as a factory.\n"
            f"Tip: seems that this is a Protocol. "
            "Please subclass it and provide the subclass."
        )


class UnsupportedGeneratorReturnTypeError(TypeError, DishkaError):
    def __init__(
            self,
            name: str,
            guess: str,
            guessed_args: str,
            *,
            is_async: bool = False,
    ) -> None:
        self.name = name
        self.guess = guess
        self.guessed_args = guessed_args
        self.is_async = is_async

    def __str__(self) -> str:
        gen_type = "async generator" if self.is_async else "generator"
        return (
            f"Unsupported return type `{self.name}` for {gen_type}. "
            f"Did you mean {self.guess}[{self.guessed_args}]?"
        )


class MissingHintsError(ValueError, DishkaError):
    def __init__(
            self,
            source: Any,
            missing_hints: Sequence[str],
            *,
            append_init: bool = False,
    ) -> None:
        self.source = source
        self.missing = missing_hints
        self.append_init = append_init

    def __str__(self) -> str:
        name = get_name(self.source, include_module=True)
        if self.append_init:
            name += ".__init__"
        missing = ", ".join(self.missing)
        return (
            f"Failed to analyze `{name}`.\n"
            f"Some parameters do not have type hints: {missing}\n"
        )


class UndefinedTypeAnalysisError(NameError, DishkaError):
    def __init__(
            self,
            source: Any,
            type_name: str,
    ) -> None:
        self.source = source
        self.type_name = type_name
        self.message = self._construct_message()
        super().__init__(self.message, name=type_name)

    def _construct_message(self) -> str:
        name = get_name(self.source, include_module=True)
        return (
            f"Failed to analyze `{name}`. \n"
            f"Type '{self.type_name}' is not defined. \n\n"
            f"If your are using `if TYPE_CHECKING` "
            f"to import '{self.type_name}' "
            f"then try removing it. \n"
            f"Or, create a separate factory with all types imported."
        )

    def __str__(self) -> str:
        return self.message


class MissingReturnHintError(ValueError, DishkaError):
    def __init__(self, source: Any) -> None:
        self.source = source

    def __str__(self) -> str:
        name = get_name(self.source, include_module=True)
        return (
            f"Failed to analyze `{name}`. \n"
            f"Missing return type hint."
        )


class NoScopeSetInProvideError(ValueError, DishkaError):
    def __init__(
            self,
            provides_name: str,
            src_name: str,
            provider_name: str,
    ) -> None:
        self.provides_name = provides_name
        self.src_name = src_name
        self.provider_name = provider_name

    def __str__(self) -> str:
        return (
            f"No scope is set for {self.provides_name}.\n"
            f"Set in provide() call for {self.src_name} or "
            f"within {self.provider_name}"
        )


class NoScopeSetInContextError(ValueError, DishkaError):
    def __init__(
            self,
            provides_name: str,
            provider_name: str,
    ) -> None:
        self.provides_name = provides_name
        self.provider_name = provider_name

    def __str__(self) -> str:
        return (
            f"No scope is set for {self.provides_name}.\n"
            f"Set in from_context() call or within {self.provider_name}"
        )


class IndependentDecoratorError(ValueError, DishkaError):
    def __init__(self, source: Callable[..., Any] | type) -> None:
        self.source = source

    def __str__(self) -> str:
        name = get_name(self.source, include_module=True)
        return (
            f"Decorator {name} does not depend on provided type.\n"
            f"Did you mean @provide instead of @decorate?"
        )



from typing import Any, get_args, get_origin

from dishka.entities.factory_type import FactoryData, FactoryType
from dishka.entities.key import DependencyKey


def _render_args(hint: Any) -> str:
    args = get_args(hint)
    return ", ".join(
        get_name(arg, include_module=False)
        for arg in args
    )


def get_name(hint: Any, *, include_module: bool) -> str:
    if hint is ...:
        return "..."
    if func := getattr(object, "__func__", None):
        return get_name(func, include_module=include_module)

    if include_module:
        module = getattr(hint, "__module__", "")
        if module == "builtins":
            module = ""
        elif module:
            module += "."
    else:
        module = ""

    name = (
        getattr(hint, "__qualname__", None) or
        getattr(hint, "__name__", None)
    )
    if name:
        if get_origin(hint):
            args = f"[{_render_args(hint)}]"
        else:
            args = ""
        return f"{module}{name}{args}"
    return str(hint)


def get_source_name(factory: FactoryData) -> str:
    source = factory.source
    if source == factory.provides.type_hint:
        return ""
    if factory.type is FactoryType.ALIAS:
        return "alias"

    return get_name(source, include_module=False)


def get_key_name(key: DependencyKey) -> str:
    return get_name(key.type_hint, include_module=True)



from faststream.__about__ import (
    __version__ as FASTSTREAM_VERSION,  # noqa: N812
)

from dishka import FromDishka

FASTSTREAM_05 = FASTSTREAM_VERSION.startswith("0.5")
FASTSTREAM_06 = FASTSTREAM_VERSION.startswith("0.6")

if FASTSTREAM_05:
    from .faststream_05 import FastStreamProvider, inject, setup_dishka
elif FASTSTREAM_06:
    from .faststream_06 import FastStreamProvider, inject, setup_dishka
else:
    raise RuntimeError(  # noqa: TRY003
        f"FastStream {FASTSTREAM_VERSION} version not supported",
    )

__all__ = (
    "FastStreamProvider",
    "FromDishka",
    "inject",
    "setup_dishka",
)



import typing
from abc import ABC, ABCMeta
from collections.abc import Iterable
from enum import Enum
from itertools import chain
from typing import (
    TYPE_CHECKING,
    Any,
    Final,
    Generic,
    Protocol,
    TypeAlias,
    TypeVar,
)

from dishka._adaptix.common import TypeHint
from dishka._adaptix.feature_requirement import (
    HAS_PY_311,
    HAS_PY_312,
    HAS_TV_TUPLE,
    HAS_UNPACK,
)
from dishka._adaptix.type_tools import (  # type: ignore[attr-defined]
    normalize_type,
)
from dishka._adaptix.type_tools.basic_utils import (
    get_type_vars_of_parametrized,
    is_generic,
    is_parametrized,
)
from dishka._adaptix.type_tools.fundamentals import (
    get_generic_args,
    get_type_vars,
    strip_alias,
)
from dishka._adaptix.type_tools.implicit_params import fill_implicit_params
from dishka.entities.provides_marker import ProvideMultiple

__all__ = ["ParentsResolver", "WithParents"]

from dishka.exception_base import DishkaError

IGNORE_TYPES: Final = (
    type,
    object,
    Enum,
    ABC,
    ABCMeta,
    Generic,
    Protocol,
    Exception,
    BaseException,
)

TypeArgsTuple: TypeAlias = tuple[TypeHint, ...]

if HAS_PY_311:

    def is_type_var_tuple(obj: TypeHint) -> bool:
        return getattr(obj, "__typing_is_unpacked_typevartuple__", False)
else:

    def is_type_var_tuple(obj: TypeHint) -> bool:
        return False


if HAS_PY_312:
    from types import (  # type: ignore[attr-defined, unused-ignore]
        get_original_bases,
    )
else:

    def get_original_bases(cls: type, /) -> tuple[Any, ...]:
        try:
            return cls.__dict__.get("__orig_bases__", cls.__bases__)
        except AttributeError:
            msg = f"Expected an instance of type, not {type(cls).__name__!r}"
            raise TypeError(msg) from None

def has_orig_bases(obj: TypeHint) -> bool:
    return hasattr(obj, "__orig_bases__")


def is_ignored_type(origin_type: TypeHint) -> bool:
    return origin_type in IGNORE_TYPES


def create_type_vars_map(obj: TypeHint) -> dict[TypeHint, TypeHint]:
    origin_obj = strip_alias(obj)
    type_vars = list(get_type_vars(origin_obj) or get_type_vars(obj))

    if not type_vars:
        return {}

    type_vars_map = {}
    arguments = list(get_generic_args(obj))
    reversed_arguments = False

    while True:
        if not type_vars:
            break

        type_var = type_vars[0]

        if isinstance(type_var, TypeVar):
            del type_vars[0]
            type_vars_map[type_var] = arguments.pop(0)
        else:
            if len(type_vars) == 1:
                if reversed_arguments:
                    arguments.reverse()
                type_vars_map[type_var] = arguments
                break
            type_vars.reverse()
            arguments.reverse()
            reversed_arguments = not reversed_arguments

    return type_vars_map


class ParentsResolver:
    def get_parents(self, child_type: TypeHint) -> list[TypeHint]:
        if is_ignored_type(strip_alias(child_type)):
            raise StartingClassIgnoredError(child_type)
        return list(self._resolve_parents(child_type))

    def _resolve_parents(self, tp: TypeHint) -> Iterable[TypeHint]:
        result = [tp]
        for parent in self._fetch_parents(tp):
            if not is_ignored_type(parent):
                result.extend(self._resolve_parents(parent))
        return result

    def _fetch_parents(self, tp: TypeHint) -> list[TypeHint]:
        if is_parametrized(tp):
            return self._get_parents_of_parametrized_generic(tp)
        if is_generic(tp):
            return self._get_parents_of_parametrized_generic(
                fill_implicit_params(tp),
            )
        return self._get_parents(tp)

    def _get_parents_of_parametrized_generic(
            self,
            parametrized_generic: TypeHint,
    ) -> list[TypeHint]:
        origin = strip_alias(parametrized_generic)
        type_var_to_actual = self._get_type_var_to_actual(
            get_type_vars(origin),
            self._unpack_args(get_generic_args(parametrized_generic)),
        )
        return [
            self._parametrize_by_dict(type_var_to_actual, tp)
            for tp in self._get_parents(origin)
        ]

    def _unpack_args(self, args: TypeArgsTuple) -> TypeArgsTuple:
        if HAS_UNPACK and any(
                strip_alias(arg) == typing.Unpack for arg in args  # type: ignore[attr-defined, unused-ignore]
        ):
            subscribed = tuple[args]  # type: ignore[valid-type]
            return tuple(arg.source for arg in normalize_type(subscribed).args)
        return args

    def _get_type_var_to_actual(
            self,
            type_vars: TypeArgsTuple,
            args: TypeArgsTuple,
    ) -> dict[TypeHint, TypeArgsTuple]:
        result = {}
        idx = 0
        for tv in type_vars:
            if HAS_TV_TUPLE and isinstance(tv, typing.TypeVarTuple):  # type: ignore[attr-defined, unused-ignore]
                tuple_len = len(args) - len(type_vars) + 1
                result[tv] = args[idx : idx + tuple_len]
                idx += tuple_len
            else:
                result[tv] = (args[idx],)
                idx += 1

        return result

    def _parametrize_by_dict(
            self,
            type_var_to_actual: dict[TypeHint, TypeArgsTuple],
            tp: TypeHint,
    ) -> TypeHint:
        params = get_type_vars_of_parametrized(tp)
        if not params:
            return tp
        return tp[
            tuple(
                chain.from_iterable(
                    type_var_to_actual[type_var] for type_var in params
                ),
            )
        ]

    def _get_parents(self, tp: TypeHint) -> list[TypeHint]:
        return [
            parent for parent in get_original_bases(tp)
            if strip_alias(parent) not in (Generic, Protocol)
        ]


if TYPE_CHECKING:
    T = TypeVar("T")
    WithParents: TypeAlias = T
else:
    class WithParents:
        def __class_getitem__(cls, item: TypeHint) -> TypeHint:
            parents = ParentsResolver().get_parents(item)
            if len(parents) > 1:
                return ProvideMultiple[tuple(parents)]
            return parents[0]


class StartingClassIgnoredError(ValueError, DishkaError):
    def __init__(self, hint: TypeHint) -> None:
        self.hint = hint

    def __str__(self) -> str:
        return f"The starting class {self.hint!r} is in ignored types"



import re

from dishka.plotter.model import Group, Node, NodeType, Renderer

HTML_TEMPLATE = """\
<html>
<head>
    <meta charset="UTF-8">
</head>
<body>

<pre class="mermaid">
{diagram}
</pre>

<script type="module">
    import mermaid
    from 'https://cdn.jsdelivr.net/npm/mermaid@11/dist/mermaid.esm.min.mjs';
    mermaid.initialize(config);
</script>
</body>
</html>

"""


class MermaidRenderer(Renderer):
    def __init__(self) -> None:
        self.nodes: dict[str, Node] = {}

    def _render_node(self, node: Node) -> str:
        if node.type is NodeType.ALIAS:
            return ""
        name = self._node_type(node) + self._escape(node.name)
        return (
            f'class {node.id}["{name}"]'
            + "{\n"
            + (f"    {node.source_name}()\n" if node.source_name else " \n")
            + "".join(
                f"    {self.nodes[dep].name}\n" for dep in node.dependencies
            )
            + "}\n"
        )

    def _escape(self, line: str) -> str:
        return re.sub(r"[^\w_.\-]", "_", line)

    def _render_node_deps(self, node: Node) -> list[str]:
        res: list[str] = []
        if node.type is NodeType.ALIAS:
            return res
        for dep in node.dependencies:
            dep_node = self.nodes[dep]
            if dep_node.type is NodeType.ALIAS:
                target = dep_node.dependencies[0]
                res.append(
                    f"{target} <.. {node.id}: "
                    f"🔗 {self._escape(dep_node.name)}",
                )
            else:
                res.append(f"{dep} <.. {node.id}")
        return res

    def _node_type(self, node: Node) -> str:
        if node.is_protocol:
            prefix = "Ⓘ "
        else:
            prefix = ""
        if node.type is NodeType.DECORATOR:
            return "🎭 " + prefix
        elif node.type is NodeType.CONTEXT:
            return "📥 " + prefix
        elif node.type is NodeType.ALIAS:
            return "🔗 " + prefix
        return "🏭 " + prefix

    def _render_group(
            self, group: Group, name_prefix: str = "",
    ) -> str:
        if not group.nodes and not group.children:
            return ""
        name = name_prefix + self._escape(group.name or "")
        res = ""
        if group.nodes:
            res = f"namespace {name} {{\n"
            for node in group.nodes:
                res += self._render_node(node) + "\n"
            res += "}\n"
        for child in group.children:
            res += self._render_group(child, name) + "\n"
        return res

    def _render_links(self, group: Group) -> str:
        res = ""
        for node in group.nodes:
            for dep_str in self._render_node_deps(node):
                res += dep_str + "\n"
        for child in group.children:
            res += self._render_links(child)
        return res

    def _fill_nodes(self, groups: list[Group]) -> None:
        for group in groups:
            for node in group.nodes:
                self.nodes[node.id] = node
            self._fill_nodes(group.children)

    def render(self, groups: list[Group]) -> str:
        self._fill_nodes(groups)

        res = "classDiagram\n"
        res += "direction LR\n"
        for group in groups:
            res += self._render_group(group)
            res += self._render_links(group)
        return HTML_TEMPLATE.format(diagram=res)



__all__ = (
    "FastStreamProvider",
    "inject",
    "setup_dishka",
)

import warnings
from collections.abc import AsyncIterator, Awaitable, Callable
from typing import (
    Any,
    ParamSpec,
    Protocol,
    TypeAlias,
    TypeVar,
    cast,
)

from faststream import BaseMiddleware, FastStream, context
from faststream.broker.core.usecase import BrokerUsecase as BrokerType
from faststream.broker.message import StreamMessage
from faststream.types import DecodedMessage
from faststream.utils.context import ContextRepo

from dishka import AsyncContainer, Provider, Scope, from_context
from dishka.integrations.base import InjectFunc, wrap_injection

_ReturnT = TypeVar("_ReturnT")
_ParamsP = ParamSpec("_ParamsP")

class FastStreamProvider(Provider):
    context = from_context(ContextRepo, scope=Scope.REQUEST)
    message = from_context(StreamMessage, scope=Scope.REQUEST)

try:
    # AsgiFastStream was introduced in FastStream 0.5.16
    from faststream.asgi import AsgiFastStream

except ImportError:
    Application: TypeAlias = FastStream  # type: ignore[no-redef,misc]

else:
    Application: TypeAlias = FastStream | AsgiFastStream  # type: ignore[no-redef,misc]

try:
    from faststream.broker.fastapi import StreamRouter
except ImportError:
    pass
else:
    Application |= StreamRouter  # type: ignore[assignment]


class ApplicationLike(Protocol):
    broker: BrokerType[Any, Any]


def setup_dishka(
    container: AsyncContainer,
    app: "Application | ApplicationLike | None" = None,
    broker: "BrokerType[Any, Any] | None" = None,
    *,
    finalize_container: bool = True,
    auto_inject: bool | InjectFunc[_ParamsP, _ReturnT] = False,
) -> None:
    """
    Setup dishka integration with FastStream.
    You must provide either app or broker.

    Args:
        container: AsyncContainer instance.
        app: FastStream Application or StreamRouter instance.
        broker: FastStream broker instance.
        finalize_container: bool. Can be used only with app.
        auto_inject: bool or custom inject decorator.
    """
    if (app and broker) or (not app and not broker):
        raise ValueError(  # noqa: TRY003
            "You must provide either app or broker "
            "to setup dishka integration.",
        )

    if finalize_container:
        if getattr(app, "after_shutdown", None):
            app.after_shutdown(container.close)  # type: ignore[union-attr]

        else:
            warnings.warn(
                "For use `finalize_container=True` "
                "you must provide `app: FastStream | AsgiFastStream` "
                "argument.",
                category=RuntimeWarning,
                stacklevel=2,
            )

    broker = broker or getattr(app, "broker", None)
    assert broker  # noqa: S101

    if getattr(broker, "add_middleware", None):
        # FastStream 0.5.6 and higher
        broker.add_middleware(DishkaMiddleware(container))

    else:
        # FastStream 0.5 - 0.5.5 versions backport
        broker._middlewares = (  # noqa: SLF001
            DishkaMiddleware(container),
            *broker._middlewares,  # noqa: SLF001
        )

        for subscriber in broker._subscribers.values():  # noqa: SLF001
            subscriber._broker_middlewares = (  # noqa: SLF001
                DishkaMiddleware(container),
                *subscriber._broker_middlewares,  # noqa: SLF001
            )

        for publisher in broker._publishers.values():  # noqa: SLF001
            publisher._broker_middlewares = (  # noqa: SLF001
                DishkaMiddleware(container),
                *publisher._broker_middlewares,  # noqa: SLF001
            )

    if auto_inject is not False:
        inject_func: InjectFunc[_ParamsP, _ReturnT]

        if auto_inject is True:
            inject_func = inject
        else:
            inject_func = auto_inject

        broker._call_decorators = (  # noqa: SLF001
            inject_func,
            *broker._call_decorators,  # noqa: SLF001
        )


def inject(func: Callable[_ParamsP, _ReturnT]) -> Callable[_ParamsP, _ReturnT]:
    return wrap_injection(
        func=func,
        is_async=True,
        container_getter=lambda *_: context.get_local("dishka"),
    )


class DishkaMiddleware:
    def __init__(self, container: AsyncContainer) -> None:
        self.container = container

    def __call__(self, *args: Any, **kwargs: Any) -> "_DishkaMiddleware":
        return _DishkaMiddleware(self.container, *args, **kwargs)


class _DishkaMiddleware(BaseMiddleware):
    def __init__(
        self,
        container: AsyncContainer,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.container = container
        super().__init__(*args, **kwargs)

    async def consume_scope(  # type: ignore[misc]
        self,
        call_next: Callable[[Any], Awaitable[Any]],
        msg: StreamMessage[Any],
    ) -> AsyncIterator[DecodedMessage]:
        async with self.container(
            {
                StreamMessage: msg,
                ContextRepo: context,
                type(msg): msg,
            },
        ) as request_container:
            with context.scope("dishka", request_container):
                return cast(
                    AsyncIterator[DecodedMessage],
                    await call_next(msg),
                )



__all__ = [
    "CONTAINER_NAME",
    "AiogramMiddlewareData",
    "AiogramProvider",
    "AutoInjectMiddleware",
    "FromDishka",
    "inject",
    "inject_handler",
    "inject_router",
    "setup_dishka",
]

import warnings
from collections.abc import Awaitable, Callable
from functools import partial
from inspect import Parameter, signature
from typing import Any, Final, NewType, ParamSpec, TypeVar, cast

from aiogram import BaseMiddleware, Router
from aiogram.dispatcher.event.handler import HandlerObject
from aiogram.types import TelegramObject

from dishka import AsyncContainer, FromDishka, Provider, Scope, from_context
from .base import InjectFunc, is_dishka_injected, wrap_injection

P = ParamSpec("P")
T = TypeVar("T")
CONTAINER_NAME: Final = "dishka_container"
AiogramMiddlewareData = NewType("AiogramMiddlewareData", dict[str, Any])


def inject(func: Callable[P, T]) -> Callable[P, T]:
    if CONTAINER_NAME in signature(func).parameters:
        additional_params = []
    else:
        additional_params = [Parameter(
            name=CONTAINER_NAME,
            annotation=AsyncContainer,
            kind=Parameter.KEYWORD_ONLY,
        )]

    return wrap_injection(
        func=func,
        is_async=True,
        additional_params=additional_params,
        container_getter=lambda args, kwargs: kwargs[CONTAINER_NAME],
    )


class AiogramProvider(Provider):
    event = from_context(TelegramObject, scope=Scope.REQUEST)
    middleware_data = from_context(AiogramMiddlewareData, scope=Scope.REQUEST)


class ContainerMiddleware(BaseMiddleware):
    def __init__(self, container: AsyncContainer) -> None:
        self.container = container

    async def __call__(
        self,
        handler: Callable[[TelegramObject, dict[str, Any]], Awaitable[Any]],
        event: TelegramObject,
        data: dict[str, Any],
    ) -> Any:
        async with self.container(
                {
                    TelegramObject: event,
                    AiogramMiddlewareData: data,
                },
        ) as sub_container:
            data[CONTAINER_NAME] = sub_container
            return await handler(event, data)


class AutoInjectMiddleware(BaseMiddleware):
    def __init__(self) -> None:
        warnings.warn(
            f"{self.__class__.__name__} is slow, "
            "use `setup_dishka` instead if you care about performance",
            UserWarning,
            stacklevel=2,
        )

    async def __call__(
        self,
        handler: Callable[[TelegramObject, dict[str, Any]], Awaitable[Any]],
        event: TelegramObject,
        data: dict[str, Any],
    ) -> Any:
        old_handler = cast(HandlerObject, data["handler"])
        if is_dishka_injected(old_handler.callback):
            return await handler(event, data)

        inject_handler(old_handler, inject)
        return await handler(event, data)


def setup_dishka(
    container: AsyncContainer,
    router: Router,
    *,
    auto_inject: bool | InjectFunc[P, T] = False,
) -> None:
    middleware = ContainerMiddleware(container)

    for observer in router.observers.values():
        observer.outer_middleware(middleware)

    if auto_inject is not False:
        inject_func: InjectFunc[P, T]

        if auto_inject is True:
            inject_func = inject
        else:
            inject_func = auto_inject

        callback = partial(
            inject_router,
            router=router,
            inject_func=inject_func,
        )
        router.startup.register(callback)


def inject_router(
    router: Router,
    inject_func: InjectFunc[P, T] = inject,
) -> None:
    """Inject dishka to the router handlers."""
    for sub_router in router.chain_tail:
        for observer in sub_router.observers.values():
            if observer.event_name == "update":
                continue

            for handler in observer.handlers:
                if not is_dishka_injected(handler.callback):
                    inject_handler(handler, inject_func)


def inject_handler(
    handler: HandlerObject,
    inject_func: InjectFunc[P, T] = inject,
) -> HandlerObject:
    """Inject dishka for callback in aiogram's handler."""
    # temp_handler is used to apply original __post_init__ processing
    # for callback object wrapped by injector
    temp_handler = HandlerObject(
        callback=inject_func(handler.callback),
        filters=handler.filters,
        flags=handler.flags,
    )

    # since injector modified callback and params,
    # we should update them in the original handler
    handler.callback = temp_handler.callback
    handler.params = temp_handler.params

    return handler



from dishka import AsyncContainer, Container
from .d2 import D2Renderer
from .mermaid import MermaidRenderer
from .transform import Transformer


def render_mermaid(container: AsyncContainer | Container) -> str:
    return MermaidRenderer().render(
        Transformer().transform(container),
    )


def render_d2(container: AsyncContainer | Container) -> str:
    return D2Renderer().render(
        Transformer().transform(container),
    )



from collections.abc import Sequence
from typing import NamedTuple

from dishka.entities.component import Component
from dishka.entities.factory_type import FactoryData
from dishka.entities.key import DependencyKey
from dishka.entities.scope import BaseScope
from dishka.text_rendering.name import get_key_name, get_source_name


class PathRenderer:
    def __init__(self, *, cycle: bool):
        self.cycle = cycle

    def _arrow_cycle(self, index: int, length: int) -> str:
        if length == 1:
            return "⥁ "
        elif index == 0:
            return "╭─>─╮ "
        elif index + 1 < length:
            return "│   ▼ "
        else:
            return "╰─<─╯ "

    def _arrow_line(self, index: int, length: int) -> str:
        if index + 1 < length:
            return "▼   "
        else:
            return "╰─> "

    def _arrow(self, index: int, length: int) -> str:
        if self.cycle:
            return self._arrow_cycle(index, length)
        return self._arrow_line(index, length)

    def _switch_arrow(self, index: int, length: int) -> str:
        if self.cycle:
            if index > 0:
                return "│   │ "
            return "      "
        return "│   "

    def _switch_filler(self) -> str:
        return " "

    def _switch(
            self, scope: BaseScope | None, component: Component | None,
    ) -> str:
        return f"◈ {scope}, component={component!r} ◈"

    def render(
            self,
            path: Sequence[FactoryData],
            last: DependencyKey | None = None,
    ) -> str:
        row_count = len(path) + bool(last)
        rows = [
            Row(
                row_num,
                self._arrow(row_num, row_count),
                [get_key_name(factory.provides), get_source_name(factory)],
                (factory.scope, factory.provides.component),
            )
            for row_num, factory in enumerate(path)
        ]
        if last:
            rows.append(
                Row(
                    row_count - 1,
                    self._arrow(row_count-1, row_count),
                    [get_key_name(last), "???"],
                    (rows[-1].dest[0], last.component),
                ),
            )
        prev_dest: tuple[BaseScope | None, Component | None] = (None, "")
        space_left = "   "
        space_between = "   "
        res = ""

        columns_count = len(rows[0].columns)
        columns_width = [
            max(len(row.columns[col_num]) for row in rows)
            for col_num in range(columns_count)
        ]
        switch_len = (
            sum(columns_width) +
            len(space_between) * (columns_count - 1)
        )
        for row in rows:
            if row.dest != prev_dest:
                res += (
                    space_left +
                    self._switch_arrow(row.num, row_count) +
                    self._switch(*row.dest).center(
                        switch_len, self._switch_filler(),
                    ) +
                    "\n"
                )
                prev_dest = row.dest
            res += (
                space_left +
                row.border_left +
                space_between.join(
                    c.ljust(cw)
                    for c, cw in zip(row.columns, columns_width, strict=False)
                ) +
                "\n"
            )
        return res


class Row(NamedTuple):
    num: int
    border_left: str
    columns: Sequence[str]
    dest: tuple[BaseScope | None, Component | None]



from dishka.exception_base import DishkaError


class InvalidInjectedFuncTypeError(DishkaError):
    def __str__(self) -> str:
        return "An async container cannot be used in a synchronous context."


class ImproperProvideContextUsageError(DishkaError):
    def __str__(self) -> str:
        return "provide_context can only be used with manage_scope=True."



__all__ = [
    "DishkaRoute",
    "DishkaSyncRoute",
    "FastapiProvider",
    "FromDishka",
    "inject",
    "inject_sync",
    "setup_dishka",
]

from collections.abc import Callable
from inspect import (
    Parameter,
    isasyncgenfunction,
    iscoroutinefunction,
    signature,
)
from typing import Annotated, Any, ParamSpec, TypeVar, get_type_hints

from fastapi import Depends, FastAPI, Request, WebSocket
from fastapi.routing import APIRoute

from dishka import (
    AsyncContainer,
    Container,
    DependencyKey,
    FromDishka,
    Provider,
    Scope,
    from_context,
)
from .base import default_parse_dependency, wrap_injection
from .starlette import ContainerMiddleware, SyncContainerMiddleware

T = TypeVar("T")
P = ParamSpec("P")


def _async_depends(dependency: DependencyKey) -> Any:
    async def dishka_depends(request: Request) -> dependency:
        return await request.state.dishka_container.get(
            dependency.type_hint,
            component=dependency.component,
        )

    return Annotated[dependency.type_hint, Depends(dishka_depends)]


def _replace_depends(
    func: Callable[P, T],
    depends_factory: Callable[[DependencyKey], Any],
) -> Callable[P, T]:
    hints = get_type_hints(func, include_extras=True)
    func_signature = signature(func)

    new_params = []
    for name, param in func_signature.parameters.items():
        hint = hints.get(name, Any)
        dep = default_parse_dependency(param, hint)
        if dep is None:
            new_params.append(param)
            continue
        new_dep = depends_factory(dep)
        hints[name] = new_dep
        new_params.append(param.replace(annotation=new_dep))
    func.__signature__ = func_signature.replace(parameters=new_params)
    func.__annotations__ = hints
    return func


def _find_request_param(func: Callable[P, T]) -> str | None:
    hints = get_type_hints(func)
    request_hint = next(
        (name for name, hint in hints.items() if hint is Request),
        None,
    )
    websocket_hint = next(
        (name for name, hint in hints.items() if hint is WebSocket),
        None,
    )
    return request_hint or websocket_hint


DISHKA_REQUEST_PARAM = Parameter(
    name="___dishka_request",
    annotation=Request,
    kind=Parameter.KEYWORD_ONLY,
)


def inject(func: Callable[P, T]) -> Callable[P, T]:
    if not iscoroutinefunction(func) and not isasyncgenfunction(func):
        return _replace_depends(func, _async_depends)
    return _wrap_fastapi_injection(func=func, is_async=True)


def inject_sync(func: Callable[P, T]) -> Callable[P, T]:
    return _wrap_fastapi_injection(func=func, is_async=False)


def _wrap_fastapi_injection(
    *,
    func: Callable[P, T],
    is_async: bool,
) -> Callable[P, T]:
    param_name = _find_request_param(func)
    if param_name:
        additional_params = []
    else:
        additional_params = [DISHKA_REQUEST_PARAM]
        param_name = DISHKA_REQUEST_PARAM.name
    return wrap_injection(
        func=func,
        is_async=is_async,
        additional_params=additional_params,
        container_getter=lambda _, p: p[param_name].state.dishka_container,
    )


class DishkaRoute(APIRoute):
    def __init__(
        self,
        path: str,
        endpoint: Callable[..., Any],
        **kwargs: Any,
    ) -> None:
        endpoint = inject(endpoint)
        super().__init__(path, endpoint, **kwargs)


class DishkaSyncRoute(APIRoute):
    def __init__(
        self,
        path: str,
        endpoint: Callable[..., Any],
        **kwargs: Any,
    ) -> None:
        endpoint = inject_sync(endpoint)
        super().__init__(path, endpoint, **kwargs)


class FastapiProvider(Provider):
    request = from_context(Request, scope=Scope.REQUEST)
    websocket = from_context(WebSocket, scope=Scope.SESSION)


def setup_dishka(container: AsyncContainer | Container, app: FastAPI) -> None:
    if isinstance(container, AsyncContainer):
        app.add_middleware(ContainerMiddleware)
    else:
        app.add_middleware(SyncContainerMiddleware)
    app.state.dishka_container = container



__all__ = [
    "DISHKA_CONTAINER_KEY",
    "AiohttpProvider",
    "FromDishka",
    "inject",
    "setup_dishka",
]

from collections.abc import Callable, Coroutine
from typing import Any, Final, TypeAlias, cast

from aiohttp import web
from aiohttp.abc import AbstractView
from aiohttp.typedefs import Handler
from aiohttp.web_app import Application
from aiohttp.web_request import Request
from aiohttp.web_response import StreamResponse

from dishka import AsyncContainer, FromDishka, Provider, Scope, from_context
from dishka.integrations.base import (
    InjectFunc,
    is_dishka_injected,
    wrap_injection,
)

DISHKA_CONTAINER_KEY: Final = web.AppKey("dishka_container", AsyncContainer)
AiohttpHandler: TypeAlias = type[AbstractView] | Callable[
    [Request], Coroutine[Any, Any, StreamResponse],
]


def inject(func: Callable[..., Any]) -> AiohttpHandler:
    return cast(
        AiohttpHandler,
        wrap_injection(
            func=func,
            is_async=True,
            container_getter=lambda p, _: p[0][DISHKA_CONTAINER_KEY],
        ),
    )


class AiohttpProvider(Provider):
    request = from_context(Request, scope=Scope.SESSION)


@web.middleware
async def container_middleware(
    request: Request,
    handler: Handler,
) -> StreamResponse:
    container = request.app[DISHKA_CONTAINER_KEY]

    if (
        request.headers.get("Upgrade") == "websocket"
        and request.headers.get("Connection") == "Upgrade"
    ):
        scope = Scope.SESSION

    else:
        scope = Scope.REQUEST

    context = {Request: request}

    async with container(context=context, scope=scope) as request_container:
        request[DISHKA_CONTAINER_KEY] = request_container  # type: ignore[index]
        return await handler(request)


def _inject_routes(
    router: web.UrlDispatcher,
    inject_func: InjectFunc[..., Any],
) -> None:
    for route in router.routes():
        _inject_route(route, inject_func)

    for resource in router.resources():
        try:
            routes = iter(resource)
        except TypeError:
            continue
        for route in routes:
            _inject_route(route, inject_func)


def _inject_route(
    route: web.AbstractRoute,
    inject_func: InjectFunc[..., Any],
) -> None:
    if not is_dishka_injected(route.handler):
        # typing.cast is used because AbstractRoute._handler
        # is Handler or Type[AbstractView]
        route._handler = cast(  # noqa: SLF001
            AiohttpHandler,
            inject_func(route.handler),
        )


async def _on_shutdown(app: web.Application) -> None:
    await app[DISHKA_CONTAINER_KEY].close()


def setup_dishka(
    container: AsyncContainer,
    app: Application,
    *,
    auto_inject: bool | InjectFunc[..., Any] = False,
    finalize_container: bool = True,
) -> None:
    app[DISHKA_CONTAINER_KEY] = container
    app.middlewares.append(container_middleware)
    if finalize_container:
        app.on_shutdown.append(_on_shutdown)

    if auto_inject is not False:
        inject_func: InjectFunc[..., Any]

        if auto_inject is True:
            inject_func = inject
        else:
            inject_func = auto_inject

        _inject_routes(app.router, inject_func)



from collections.abc import Sequence

from dishka.entities.factory_type import FactoryData
from dishka.entities.key import DependencyKey
from .name import get_key_name, get_name, get_source_name


def render_suggestions_for_missing(
    requested_for: FactoryData | None,
    requested_key: DependencyKey,
    suggest_other_scopes: Sequence[FactoryData],
    suggest_other_components: Sequence[FactoryData],
    suggest_abstract_factories: Sequence[FactoryData],
    suggest_concrete_factories: Sequence[FactoryData],
) -> str:
    suggestion = ""
    if suggest_other_scopes:
        scopes = " or ".join(
            str(factory.scope)
            for factory in suggest_other_scopes
        )
        if requested_for:
            srcname = get_name(requested_for.source, include_module=False)
            suggestion += f"\n * Try changing scope of `{srcname}` to {scopes}"
        else:
            suggestion += f"\n * Check if you forgot enter {scopes}"

    if suggest_other_components:
        dep_name = get_name(requested_key.type_hint, include_module=True)
        components_str = " or ".join(
            f"Annotated[{dep_name}, FromComponent({f.provides.component!r})]"
            for f in suggest_other_components
        )
        suggestion += f"\n * Try using {components_str}"

    if suggest_abstract_factories:
        abstract_names = ""
        for factory in suggest_abstract_factories:
            source_name = get_source_name(factory)
            provides_name = get_key_name(factory.provides)
            abstract_names += "(" + provides_name
            if source_name:
                abstract_names += ", " + source_name
            abstract_names += ");"

        suggestion += "\n * Try use `AnyOf` "
        suggestion += "or changing the requested dependency to a more abstract"
        suggestion += ". Found factories for more abstract dependencies: "
        suggestion += abstract_names

    if suggest_concrete_factories:
        concreate_names = ""
        for factory in suggest_concrete_factories:
            source_name = get_source_name(factory)
            provides_name = get_key_name(factory.provides)
            concreate_names += "(" + provides_name
            if source_name:
                concreate_names += ", " + source_name
            concreate_names += ");"

        deb_name = get_name(requested_key.type_hint, include_module=True)

        suggestion += "\n * Try use `WithParents` "
        suggestion += "or changing `provides` to "
        suggestion += f"`{deb_name}`"

        suggestion += ". Found factories for more concrete dependencies: "
        suggestion += concreate_names

    return suggestion



from typing import Any

from dishka.dependency_source import (
    Alias,
    CompositeDependencySource,
    ContextVariable,
    context_stub,
)
from dishka.entities.component import DEFAULT_COMPONENT
from dishka.entities.key import DependencyKey
from dishka.entities.scope import BaseScope
from dishka.entities.type_alias_type import (
    is_type_alias_type,
    unwrap_type_alias,
)


def from_context(
        provides: Any,
        *,
        scope: BaseScope | None = None,
        override: bool = False,
) -> CompositeDependencySource:
    composite = CompositeDependencySource(origin=context_stub)
    composite.dependency_sources.append(
        ContextVariable(
            scope=scope,
            override=override,
            provides=DependencyKey(
                type_hint=provides,
                component=DEFAULT_COMPONENT,
            ),
        ),
    )

    if is_type_alias_type(provides):
        base_type = unwrap_type_alias(provides)
        composite.dependency_sources.append(
            Alias(
                source=DependencyKey(provides, DEFAULT_COMPONENT),
                provides=DependencyKey(base_type, DEFAULT_COMPONENT),
                cache=True,
                override=override,
            ),
        )
    return composite



__all__ = [
    "DishkaAioInterceptor",
    "DishkaInterceptor",
    "FromDishka",
    "GrpcioProvider",
    "inject",
]

from collections.abc import Awaitable, Callable, Iterator
from contextvars import ContextVar
from inspect import isasyncgenfunction, iscoroutine, iscoroutinefunction
from typing import Any, ParamSpec, TypeVar

from google.protobuf.message import Message
from grpc import (
    HandlerCallDetails,
    RpcMethodHandler,
    ServerInterceptor,
    ServicerContext,
    stream_stream_rpc_method_handler,
    stream_unary_rpc_method_handler,
    unary_stream_rpc_method_handler,
    unary_unary_rpc_method_handler,
)
from grpc.aio import ServerInterceptor as AioServerInterceptor

from dishka import (
    AsyncContainer,
    Container,
    FromDishka,
    Provider,
    Scope,
    from_context,
)
from dishka.integrations.base import wrap_injection

P = ParamSpec("P")
RT = TypeVar("RT")

_dishka_scoped_container = ContextVar("_dishka_scoped_container")


def inject(func: Callable[P, RT]) -> Callable[P, RT]:
    return wrap_injection(
        func=func,
        is_async=iscoroutinefunction(func) or isasyncgenfunction(func),
        container_getter=lambda _, __: _dishka_scoped_container.get(),
    )


class GrpcioProvider(Provider):
    message = from_context(Message, scope=Scope.REQUEST)
    servicer_context = from_context(ServicerContext, scope=Scope.SESSION)


class DishkaInterceptor(ServerInterceptor):   # type: ignore[misc]
    def __init__(self, container: Container) -> None:
        self._container = container

    def intercept_service(
        self,
        continuation: Callable[
            [HandlerCallDetails],
            RpcMethodHandler,
        ],
        handler_call_details: HandlerCallDetails,
    ) -> RpcMethodHandler:
        rpc_handler = continuation(handler_call_details)

        def unary_unary_behavior(
            request: Message,
            context: ServicerContext,
        ) -> Any:
            context_ = {
                Message: request,
                ServicerContext: context,
            }
            with self._container(context=context_) as container:
                _dishka_scoped_container.set(container)
                return rpc_handler.unary_unary(request, context)

        def stream_unary_behavior(
            request_iterator: Iterator[Message],
            context: ServicerContext,
        ) -> Any:
            context_ = {ServicerContext: context}
            with self._container(
                context=context_,
                scope=Scope.SESSION,
            ) as container:
                _dishka_scoped_container.set(container)
                return rpc_handler.stream_unary(
                    request_iterator, context,
                )

        def unary_stream_behavior(
            request: Message,
            context: ServicerContext,
        ) -> Any:
            context_ = {
                Message: request,
                ServicerContext: context,
            }
            with self._container(context=context_) as container:
                _dishka_scoped_container.set(container)
                yield from rpc_handler.unary_stream(request, context)

        def stream_stream_behavior(
            request_iterator: Iterator[Message],
            context: ServicerContext,
        ) -> Any:
            context_ = {ServicerContext: context}
            with self._container(
                context=context_,
                scope=Scope.SESSION,
            ) as container:
                _dishka_scoped_container.set(container)
                yield from rpc_handler.stream_stream(request_iterator, context)

        if rpc_handler.unary_unary:
            return unary_unary_rpc_method_handler(
                unary_unary_behavior,
                rpc_handler.request_deserializer,
                rpc_handler.response_serializer,
            )
        elif rpc_handler.stream_unary:
            return stream_unary_rpc_method_handler(
                stream_unary_behavior,
                rpc_handler.request_deserializer,
                rpc_handler.response_serializer,
            )
        elif rpc_handler.unary_stream:
            return unary_stream_rpc_method_handler(
                unary_stream_behavior,
                rpc_handler.request_deserializer,
                rpc_handler.response_serializer,
            )
        elif rpc_handler.stream_stream:
            return stream_stream_rpc_method_handler(
                stream_stream_behavior,
                rpc_handler.request_deserializer,
                rpc_handler.response_serializer,
            )

        return rpc_handler


class DishkaAioInterceptor(AioServerInterceptor):  # type: ignore[misc]
    def __init__(self, container: AsyncContainer) -> None:
        self._container = container

    async def intercept_service(  # noqa: C901
        self,
        continuation: Callable[
            [HandlerCallDetails],
            Awaitable[RpcMethodHandler],
        ],
        handler_call_details: HandlerCallDetails,
    ) -> RpcMethodHandler:
        rpc_handler = await continuation(handler_call_details)

        async def unary_unary_behavior(
            request: Message,
            context: ServicerContext,
        ) -> Any:
            context_ = {
                Message: request,
                ServicerContext: context,
            }
            async with self._container(context=context_) as container:
                _dishka_scoped_container.set(container)
                return await rpc_handler.unary_unary(request, context)

        async def stream_unary_behavior(
            request_iterator: Iterator[Message],
            context: ServicerContext,
        ) -> Any:
            context_ = {ServicerContext: context}
            async with self._container(
                context=context_,
                scope=Scope.SESSION,
            ) as container:
                _dishka_scoped_container.set(container)
                return await rpc_handler.stream_unary(
                    request_iterator, context,
                )

        async def unary_stream_behavior(
            request: Message,
            context: ServicerContext,
        ) -> Any:
            context_ = {
                Message: request,
                ServicerContext: context,
            }
            async with self._container(
                context=context_,
                scope=Scope.REQUEST,
            ) as container:
                _dishka_scoped_container.set(container)
                res = rpc_handler.unary_stream(request, context)
                if iscoroutine(res):
                    return await res
                async for result in res:
                    await context.write(result)
                return None

        async def stream_stream_behavior(
            request_iterator: Iterator[Message],
            context: ServicerContext,
        ) -> Any:
            context_ = {ServicerContext: context}
            async with self._container(
                context=context_,
                scope=Scope.SESSION,
            ) as container:
                _dishka_scoped_container.set(container)
                res = rpc_handler.stream_stream(request_iterator, context)
                if iscoroutine(res):
                    return await res
                async for result in res:
                    await context.write(result)
                return None

        if rpc_handler.unary_unary:
            return unary_unary_rpc_method_handler(
                unary_unary_behavior,
                rpc_handler.request_deserializer,
                rpc_handler.response_serializer,
            )
        elif rpc_handler.stream_unary:
            return stream_unary_rpc_method_handler(
                stream_unary_behavior,
                rpc_handler.request_deserializer,
                rpc_handler.response_serializer,
            )
        elif rpc_handler.unary_stream:
            return unary_stream_rpc_method_handler(
                unary_stream_behavior,
                rpc_handler.request_deserializer,
                rpc_handler.response_serializer,
            )
        elif rpc_handler.stream_stream:
            return stream_stream_rpc_method_handler(
                stream_stream_behavior,
                rpc_handler.request_deserializer,
                rpc_handler.response_serializer,
            )

        return rpc_handler



import ast
import inspect
from collections.abc import Callable, Sequence
from types import ModuleType


def make_fragments_collector(*, typing_modules: Sequence[str]) -> Callable[[ast.Module], list[ast.stmt]]:
    def check_condition(expr: ast.expr) -> bool:
        # searches for `TYPE_CHECKING`
        if (
            isinstance(expr, ast.Name)
            and isinstance(expr.ctx, ast.Load)
            and expr.id == "TYPE_CHECKING"
        ):
            return True

        # searches for `typing.TYPE_CHECKING`
        if (  # noqa: SIM103
            isinstance(expr, ast.Attribute)
            and expr.attr == "TYPE_CHECKING"
            and isinstance(expr.ctx, ast.Load)
            and isinstance(expr.value, ast.Name)
            and expr.value.id in typing_modules
            and isinstance(expr.value.ctx, ast.Load)
        ):
            return True
        return False

    def collect_type_checking_only_fragments(module: ast.Module) -> list[ast.stmt]:
        fragments = []
        for stmt in module.body:
            if isinstance(stmt, ast.If) and not stmt.orelse and check_condition(stmt.test):
                fragments.extend(stmt.body)

        return fragments

    return collect_type_checking_only_fragments


default_collector = make_fragments_collector(typing_modules=["typing"])


def exec_type_checking(
    module: ModuleType,
    *,
    collector: Callable[[ast.Module], list[ast.stmt]] = default_collector,
) -> None:
    """This function scans module source code,
    collects fragments under ``if TYPE_CHECKING`` and ``if typing.TYPE_CHECKING``
    and executes them in the context of module.
    After these, all imports and type definitions became available at runtime for analysis.

    By default, it ignores ``if`` with ``else`` branch.

    :param module: A module for processing
    :param collector: A function collecting code fragments to execute
    """
    source = inspect.getsource(module)
    fragments = collector(ast.parse(source))
    code = compile(ast.Module(fragments, type_ignores=[]), f"<exec_type_checking of {module}>", "exec")
    namespace = module.__dict__.copy()
    exec(code, namespace)  # noqa: S102
    for k, v in namespace.items():
        if not hasattr(module, k):
            setattr(module, k, v)



import re

from dishka.plotter.model import Group, Node, NodeType, Renderer


class D2Renderer(Renderer):
    def __init__(self) -> None:
        self.nodes: dict[str, Node] = {}
        self.full_ids: dict[str, str] = {}

    def _escape(self, line: str) -> str:
        return re.sub(r"[^\w_.\-]", "_", line)

    def _render_node(self, node: Node) -> str:
        if node.type is NodeType.ALIAS:
            return ""
        name = self._node_type(node) + self._escape(node.name)
        if node.source_name:
            source = f'    "{self._escape(node.source_name)}()": ""\n'
        else:
            source = ""

        res = f'{node.id}: "{name}"' + "{\n"
        res += "    shape: class\n"
        res += source
        for dep in node.dependencies:
            dep_name = self._escape(self.nodes[dep].name)
            res += f"    {dep_name}\n"
        res += "}\n"
        return res

    def _render_node_deps(self, node: Node) -> list[str]:
        res: list[str] = []
        if node.type is NodeType.ALIAS:
            return res
        node_id = self.full_ids[node.id]
        for dep in node.dependencies:
            dep_node = self.nodes[dep]
            if dep_node.type is NodeType.ALIAS:
                target = self.full_ids[dep_node.dependencies[0]]
                res.append(
                    f"{target} --> {node_id}: "
                    f"🔗 {self._escape(dep_node.name)}",
                )
            else:
                target = self.full_ids[dep]
                res.append(f"{target} --> {node_id}")
        return res

    def _node_type(self, node: Node) -> str:
        if node.is_protocol:
            prefix = "Ⓘ "
        else:
            prefix = ""
        if node.type is NodeType.DECORATOR:
            return "🎭 " + prefix
        elif node.type is NodeType.CONTEXT:
            return "📥 " + prefix
        elif node.type is NodeType.ALIAS:
            return "🔗 " + prefix
        return "🏭 " + prefix

    def _render_group(
            self, group: Group, name_prefix: str = "",
    ) -> str:
        name = name_prefix + self._escape(group.name or "")
        res = f'{group.id}: "{name}"' + "{\n"
        for node in group.nodes:
            res += self._render_node(node) + "\n"
        for child in group.children:
            res += self._render_group(child, name) + "\n"
        res += "}\n"
        return res

    def _render_links(self, group: Group) -> str:
        res = ""
        for node in group.nodes:
            for dep_str in self._render_node_deps(node):
                res += dep_str + "\n"
        for child in group.children:
            res += self._render_links(child)
        return res

    def _fill_names(self, groups: list[Group], prefix: str = "") -> None:
        for group in groups:
            if prefix:
                id_prefix = prefix + "." + group.id
            else:
                id_prefix = group.id
            self.full_ids[group.id] = id_prefix
            for node in group.nodes:
                self.nodes[node.id] = node
                self.full_ids[node.id] = id_prefix + "." + node.id
            self._fill_names(group.children, prefix=id_prefix)

    def render(self, groups: list[Group]) -> str:
        self._fill_names(groups)

        res = ""
        for group in groups:
            res += self._render_group(group)
            res += self._render_links(group)
        return res



import warnings

from dishka.entities.validation_settings import (
    DEFAULT_VALIDATION,
    STRICT_VALIDATION,
    ValidationSettings,
)

warnings.warn(
    "`dishka.entities.validation_settigs`"
    " is deprecated and will be removed in 2.0"
    ", use `from dishka import ...` instead.",
    DeprecationWarning,
    stacklevel=2,
)

__all__ = [
    "DEFAULT_VALIDATION",
    "STRICT_VALIDATION",
    "ValidationSettings",
]



__all__ = [
    "DishkaRouter",
    "FromDishka",
    "LitestarProvider",
    "inject",
    "inject_websocket",
    "setup_dishka",
]

from collections.abc import Callable
from functools import wraps
from inspect import Parameter
from typing import ParamSpec, TypeVar, get_type_hints

from litestar import Controller, Litestar, Request, Router, WebSocket
from litestar.enums import ScopeType
from litestar.handlers import (
    BaseRouteHandler,
    HTTPRouteHandler,
    WebsocketListener,
)
from litestar.handlers.websocket_handlers import WebsocketListenerRouteHandler
from litestar.handlers.websocket_handlers._utils import ListenerHandler
from litestar.routes import BaseRoute
from litestar.types import (
    ASGIApp,
    ControllerRouterHandler,
    Receive,
    Scope,
    Send,
)

from dishka import AsyncContainer, FromDishka, Provider, from_context
from dishka import Scope as DIScope
from dishka.integrations.base import wrap_injection

P = ParamSpec("P")
T = TypeVar("T")


def inject(func: Callable[P, T]):
    return _inject_wrapper(func, "request", Request)


def inject_websocket(func: Callable[P, T]):
    return _inject_wrapper(func, "socket", WebSocket)


def _inject_wrapper(
        func: Callable[P, T],
        param_name: str,
        param_annotation: type[Request | WebSocket],
):
    hints = get_type_hints(func)

    request_param = next(
        (name for name in hints if name == param_name),
        None,
    )

    if request_param:
        additional_params = []
    else:
        additional_params = [Parameter(
            name=param_name,
            annotation=param_annotation,
            kind=Parameter.KEYWORD_ONLY,
        )]

    return wrap_injection(
        func=func,
        is_async=True,
        additional_params=additional_params,
        container_getter=lambda _, r: r[param_name].state.dishka_container,
    )


def _inject_based_on_handler_type(
    value: BaseRouteHandler,
) -> BaseRouteHandler:
    if isinstance(value, HTTPRouteHandler):
        value._fn = inject(value._fn)  # noqa: SLF001

    if isinstance(value, WebsocketListenerRouteHandler) and isinstance(
        value._fn,  # noqa: SLF001
        ListenerHandler,
    ):
        value = value(inject_websocket(value._fn._fn))  # noqa: SLF001

    return value


def _inject_route_handlers(
    get_route_handlers: Callable[P, list[BaseRouteHandler]],
) -> Callable[P, list[BaseRouteHandler]]:
    @wraps(get_route_handlers)
    def _wrapper(*args: P.args, **kwargs: P.kwargs) -> list[BaseRouteHandler]:
        return [
            _inject_based_on_handler_type(route)
            for route in get_route_handlers(*args, **kwargs)
        ]

    return _wrapper


def _resolve_value(
    router: Router,
    value: ControllerRouterHandler,
) -> ControllerRouterHandler:
    if isinstance(value, Router):
        return value

    if isinstance(value, BaseRouteHandler):
        return _inject_based_on_handler_type(value)

    if isinstance(value, type):
        if issubclass(value, Controller):
            value.get_route_handlers = _inject_route_handlers(  # type: ignore[method-assign]
                value.get_route_handlers,
            )
        if issubclass(value, WebsocketListener):
            return _inject_based_on_handler_type(value(router).to_handler())

    return value


class DishkaRouter(Router):
    __slots__ = ()

    def register(self, value: ControllerRouterHandler) -> list[BaseRoute]:
        return super().register(_resolve_value(self, value))


class LitestarProvider(Provider):
    request = from_context(Request, scope=DIScope.REQUEST)
    socket = from_context(WebSocket, scope=DIScope.SESSION)


def make_add_request_container_middleware(app: ASGIApp) -> ASGIApp:
    async def middleware(scope: Scope, receive: Receive, send: Send) -> None:
        if scope.get("type") not in (ScopeType.HTTP, ScopeType.WEBSOCKET):
            await app(scope, receive, send)
            return

        if scope.get("type") == ScopeType.HTTP:
            request = Request(scope)  # type: ignore[var-annotated]
            context = {Request: request}
            di_scope = DIScope.REQUEST

        else:
            request = WebSocket(scope)
            context = {WebSocket: request}
            di_scope = DIScope.SESSION

        async with request.app.state.dishka_container(
            context, scope=di_scope,
        ) as request_container:
            request.state.dishka_container = request_container
            await app(scope, receive, send)

    return middleware


def setup_dishka(container: AsyncContainer, app: Litestar) -> None:
    app.asgi_handler = make_add_request_container_middleware(
        app.asgi_handler,
    )
    app.state.dishka_container = container



from collections.abc import Sequence
from typing import get_args, get_origin

from dishka.dependency_source import (
    Alias,
    Decorator,
    DependencySource,
    Factory,
)
from dishka.entities.key import hint_to_dependency_key
from dishka.entities.provides_marker import ProvideMultiple


def unpack_factory(factory: Factory) -> Sequence[DependencySource]:
    if get_origin(factory.provides.type_hint) is not ProvideMultiple:
        return [factory]

    provides_first, *provides_others = get_args(factory.provides.type_hint)

    res: list[DependencySource] = [
        Alias(
            provides=hint_to_dependency_key(
                provides_other,
            ).with_component(factory.provides.component),
            source=hint_to_dependency_key(
                provides_first,
            ).with_component(factory.provides.component),
            cache=factory.cache,
            override=factory.override,
        )
        for provides_other in provides_others
    ]
    res.append(
        Factory(
            dependencies=factory.dependencies,
            kw_dependencies=factory.kw_dependencies,
            type_=factory.type,
            source=factory.source,
            scope=factory.scope,
            is_to_bind=factory.is_to_bind,
            cache=factory.cache,
            override=factory.override,
            provides=hint_to_dependency_key(
                provides_first,
            ).with_component(factory.provides.component),
        ),
    )
    return res


def unpack_decorator(decorator: Decorator) -> Sequence[DependencySource]:
    if get_origin(decorator.provides.type_hint) is not ProvideMultiple:
        return [decorator]

    return [
        Decorator(
            factory=decorator.factory,
            provides=hint_to_dependency_key(
                provides,
            ).with_component(decorator.provides.component),
        )
        for provides in get_args(decorator.provides.type_hint)
    ]


def unpack_alias(alias: Alias) -> Sequence[DependencySource]:
    if get_origin(alias.provides.type_hint) is not ProvideMultiple:
        return [alias]

    return [
        Alias(
            provides=hint_to_dependency_key(
                provides,
            ).with_component(alias.provides.component),
            source=alias.source,
            cache=alias.cache,
            override=alias.override,
        )
        for provides in get_args(alias.provides.type_hint)
    ]



__all__ = [
    "BaseProvider",
    "Provider",
    "ProviderWrapper",
    "alias",
    "decorate",
    "from_context",
    "make_root_context_provider",
    "provide",
    "provide_all",
]

from .base_provider import BaseProvider, ProviderWrapper
from .make_alias import alias
from .make_context_var import from_context
from .make_decorator import decorate
from .make_factory import provide, provide_all
from .provider import Provider
from .root_context import make_root_context_provider



from typing import Any

from dishka.dependency_source import (
    Alias,
    CompositeDependencySource,
    ensure_composite,
)
from dishka.entities.component import Component
from dishka.entities.key import hint_to_dependency_key
from .unpack_provides import unpack_alias


def alias(
        source: Any,
        *,
        provides: Any | None = None,
        cache: bool = True,
        component: Component | None = None,
        override: bool = False,
) -> CompositeDependencySource:
    if component is provides is None:
        raise ValueError(  # noqa: TRY003
            "Either component or provides must be set in alias",
        )
    if provides is None:
        provides = source

    composite = ensure_composite(source)
    alias_instance = Alias(
        source=hint_to_dependency_key(source).with_component(component),
        provides=hint_to_dependency_key(provides),
        cache=cache,
        override=override,
    )
    composite.dependency_sources.extend(unpack_alias(alias_instance))
    return composite



__all__ = ["CONTAINER_NAME", "inject"]

from collections.abc import Awaitable, Callable
from typing import (
    Any,
    Concatenate,
    Final,
    ParamSpec,
    TypeVar,
    cast,
    overload,
)

from aiogram_dialog import Data, DialogManager
from aiogram_dialog.api.entities import ChatEvent
from aiogram_dialog.api.internal import Widget
from aiogram_dialog.widgets.common import ManagedWidget

from dishka import AsyncContainer
from dishka.integrations.base import wrap_injection

CONTAINER_NAME: Final = "dishka_container"
ON_DIALOG_EVENT_LEN_ARGS: Final = 2

_ParamsP = ParamSpec("_ParamsP")
_ReturnT = TypeVar("_ReturnT", bound=Awaitable[Any])

_EventT = TypeVar("_EventT", bound=ChatEvent)
_OnDialogEventData = TypeVar("_OnDialogEventData", bound=Data)
_OnProccessResultEventResultT = TypeVar("_OnProccessResultEventResultT")

_WidgetItemT = TypeVar("_WidgetItemT")
_WidgetT = TypeVar("_WidgetT", bound=Widget)
_ManagedWidgetT = TypeVar("_ManagedWidgetT", bound=ManagedWidget[Any])
_DialogManagerT = TypeVar("_DialogManagerT", bound=DialogManager)


def _container_getter(
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> AsyncContainer:
    if not args:
        container = kwargs[CONTAINER_NAME]
    elif len(args) == ON_DIALOG_EVENT_LEN_ARGS:
        container = args[-1].middleware_data[CONTAINER_NAME]
    else:
        container = args[2].middleware_data[CONTAINER_NAME]
    # typing.cast is used because middleware_data is not typed
    return cast(AsyncContainer, container)


# overload for callback with 4 arguments and managed widget
@overload
def inject(
    func: Callable[
        Concatenate[
            _EventT,
            _ManagedWidgetT,
            _DialogManagerT,
            _WidgetItemT,
            _ParamsP,
        ],
        _ReturnT,
    ],
) -> Callable[
    [_EventT, _ManagedWidgetT, _DialogManagerT, _WidgetItemT],
    _ReturnT,
]:
    ...

# overload for callback with 3 arguments
@overload
def inject(
    func: Callable[
        Concatenate[
            _EventT,
            _WidgetT,
            _DialogManagerT,
            _ParamsP,
        ],
        _ReturnT,
    ],
) -> Callable[[_EventT, _WidgetT, _DialogManagerT], _ReturnT]:
    ...

# overload for on result event
# `type: ignore[overload-overlap]` is used because mypy is dumb
@overload
def inject(  # type: ignore[overload-overlap]
    func: Callable[
        Concatenate[
            _OnDialogEventData,
            _OnProccessResultEventResultT,
            _DialogManagerT,
            _ParamsP,
        ],
        _ReturnT,
    ],
) -> Callable[
    [_OnDialogEventData, _OnProccessResultEventResultT, _DialogManagerT],
    _ReturnT,
]:
    ...

# overload for on dialog event
@overload
def inject(
    func: Callable[
        Concatenate[
            _OnDialogEventData,
            _DialogManagerT,
            _ParamsP,
        ],
        _ReturnT,
    ],
) -> Callable[[_OnDialogEventData, _DialogManagerT], _ReturnT]:
    ...


# overload for getter and custom callbacks
@overload
def inject(func: Callable[..., _ReturnT]) -> Callable[..., _ReturnT]:
    ...


def inject(func: Callable[..., _ReturnT]) -> Callable[..., _ReturnT]:
    return wrap_injection(
        func=func,
        is_async=True,
        container_getter=_container_getter,
    )



__all__ = (
    "FastStreamProvider",
    "inject",
    "setup_dishka",
)

import warnings
from collections.abc import AsyncIterator, Awaitable, Callable
from inspect import Parameter
from typing import (
    Annotated,
    Any,
    ParamSpec,
    Protocol,
    TypeAlias,
    TypeVar,
    cast,
    get_type_hints,
)

from faststream import BaseMiddleware, Context, FastStream
from faststream._internal.basic_types import DecodedMessage
from faststream._internal.broker import BrokerUsecase as BrokerType
from faststream._internal.context import ContextRepo
from faststream.asgi import AsgiFastStream
from faststream.message import StreamMessage

from dishka import AsyncContainer, Provider, Scope, from_context
from dishka.integrations.base import InjectFunc, wrap_injection

_ReturnT = TypeVar("_ReturnT")
_ParamsP = ParamSpec("_ParamsP")

class FastStreamProvider(Provider):
    context = from_context(ContextRepo, scope=Scope.REQUEST)
    message = from_context(StreamMessage, scope=Scope.REQUEST)


Application: TypeAlias = FastStream | AsgiFastStream  # type: ignore[no-redef,misc]

try:
    # import works only if fastapi is installed
    from faststream._internal.fastapi import Context, StreamRouter

except ImportError:
    from faststream import Context

else:
    Application |= StreamRouter  # type: ignore[assignment]


class ApplicationLike(Protocol):
    broker: BrokerType[Any, Any]


def setup_dishka(
    container: AsyncContainer,
    app: "Application | ApplicationLike | None" = None,
    broker: "BrokerType[Any, Any] | None" = None,
    *,
    finalize_container: bool = False,
    auto_inject: bool | InjectFunc[_ParamsP, _ReturnT] = False,
) -> None:
    """Setup dishka integration with FastStream.

    You must provide either app or broker.

    Args:
        container: AsyncContainer instance.
        app: FastStream Application or StreamRouter instance.
        broker: FastStream broker instance.
        finalize_container: bool. Can be used only with app.
        auto_inject: bool or custom inject func.
    """
    if (app and broker) or (not app and not broker):
        raise ValueError(  # noqa: TRY003
            "You must provide either app or broker "
            "to setup dishka integration.",
        )

    if finalize_container:
        if getattr(app, "after_shutdown", None):
            app.after_shutdown(container.close)  # type: ignore[union-attr]

        else:
            warnings.warn(
                "For use `finalize_container=True` "
                "you must provide `app: FastStream | AsgiFastStream` "
                "argument.",
                category=RuntimeWarning,
                stacklevel=2,
            )

    broker: BrokerType = broker or getattr(app, "broker", None)
    assert broker  # noqa: S101

    broker.insert_middleware(DishkaMiddleware(container))

    if auto_inject is not False:
        inject_func: InjectFunc[_ParamsP, _ReturnT]

        if auto_inject is True:
            inject_func = inject
        else:
            inject_func = auto_inject

        broker.config.fd_config.call_decorators = (
            inject_func,
            *broker.config.fd_config.call_decorators,
        )


class DishkaMiddleware:
    def __init__(self, container: AsyncContainer) -> None:
        self.container = container

    def __call__(self, *args: Any, **kwargs: Any) -> "_DishkaMiddleware":
        return _DishkaMiddleware(self.container, *args, **kwargs)


class _DishkaMiddleware(BaseMiddleware):
    def __init__(
        self,
        container: AsyncContainer,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.container = container
        super().__init__(*args, **kwargs)

    async def consume_scope(  # type: ignore[misc]
        self,
        call_next: Callable[[Any], Awaitable[Any]],
        msg: StreamMessage[Any],
    ) -> AsyncIterator[DecodedMessage]:
        async with self.container(
            {
                StreamMessage: msg,
                ContextRepo: self.context,
                type(msg): msg,
            },
        ) as request_container:
            with self.context.scope("dishka", request_container):
                return cast(
                    AsyncIterator[DecodedMessage],
                    await call_next(msg),
                )


def inject(func: Callable[_ParamsP, _ReturnT]) -> Callable[_ParamsP, _ReturnT]:
    param_name = _find_context_param(func)
    if param_name:
        additional_params = []
    else:
        additional_params = [DISHKA_CONTEXT_PARAM]
        param_name = DISHKA_CONTEXT_PARAM.name

    return wrap_injection(
        func=func,
        is_async=True,
        additional_params=additional_params,
        container_getter=lambda _, p: p[param_name].get_local("dishka"),
    )


def _find_context_param(func: Callable[_ParamsP, _ReturnT]) -> str | None:
    hints = get_type_hints(func)
    return next(
        (name for name, hint in hints.items() if hint is ContextRepo),
        None,
    )


DISHKA_CONTEXT_PARAM = Parameter(
    name="___dishka_context",
    annotation=Annotated[ContextRepo, Context("context")],
    kind=Parameter.KEYWORD_ONLY,
)
